using System;
using System.Collections.Generic;
using ETHZ.C0284H.Calculator.Exceptions.Calc;

namespace ETHZ.C0284H.Calculator.CalculatorNode.Visitor
{
    /// <summary>
    /// A specific visitor meant to evaluate expressions in rpn notation
    /// </summary>
    public class Evaluator : IVisitor
    {
        // Node stack to evaluate.
        private readonly Stack<INode> _innerStack;

        // Result of last visiting.
        private double _result;

        /// <summary>
        /// Costruct an evaluator given a stack of operands and operators.
        /// </summary>
        /// <param name="st">
        /// A <see cref="Stack{INode}"/>
        /// </param>
        public Evaluator(Stack<INode> st)
        {
            _innerStack = st;
        }

        /// <summary>
        /// Result of the last visiting.
        /// </summary>
        public double Result
        {
            get { return _result; }
            protected set { _result = value; }
        }

        /// <summary>
        /// Visit the operand on the stack, update Result
        /// </summary>
        /// <param name="op">
        /// A <see cref="Operand"/>
        /// </param>
        public void VisitOperand(Operand op)
        {
            if (op != null)
            {
                Result = op.OperandValue;
            }
        }

        /// <summary>
        /// Visit operator "+", update Result
        /// </summary>
        /// <param name="op">
        /// A <see cref="PlusOperator"/>
        /// </param>
        public void VisitPlusOperator(PlusOperator op)
        {
            double d1;
            double d2;
            d1 = getNextOperand();
            d2 = getNextOperand();
            Result = d1 + d2;
        }
        /// <summary>
        /// Visit operator "-", update Result
        /// </summary>
        /// <param name="opt">
        /// A <see cref="MinusOperator"/>
        /// </param>
        public void VisitMinusOperator(MinusOperator opt)
        {
            double d1;
            double d2;
            d1 = getNextOperand();
            d2 = getNextOperand();
            Result = d2 - d1;
        }
        /// <summary>
        /// Visit operator "*", update Result
        /// </summary>
        /// <param name="op">
        /// A <see cref="TimesOperator"/>
        /// </param>
        public void VisitTimesOperator(TimesOperator op)
        {
            double d1;
            double d2;
            d1 = getNextOperand();
            d2 = getNextOperand();
            Result = d2 * d1;
        }
        /// <summary>
        /// Visit operator "/", update Result
        /// </summary>
        /// <param name="op">
        /// A <see cref="DivideByOperator"/>
        /// </param>
        public void VisitDivideByOperator(DivideByOperator op)
        {
            double d1;
            double d2;
            d1 = getNextOperand();
            d2 = getNextOperand();
            Result = d2 / d1;
        }

        // Evaluate next operand on the top of stack.
        private double getNextOperand()
        {
            INode n;
            try
            {
                n = _innerStack.Pop();
                n.Accept(this);
                return Result;
            }
            catch (Exception ese)
            {
                Console.Error.Write(ese.Message);
                throw new IllegalNumberOfOperandsException("\nUnexpected number of operands on stack. Evaluation impossible: too few elements on stack.");
            }
        }
    }
}
