using System;
using System.Collections.Generic;
using System.Text;
using ETHZ.C0284H.Calculator.CalculatorNode;
using ETHZ.C0284H.Calculator.CalculatorNode.Visitor;
using ETHZ.C0284H.Calculator.Exceptions;
using ETHZ.C0284H.Calculator.Exceptions.Calc;

namespace ETHZ.C0284H.Calculator
{
	/// <summary>
	/// Models a simple RPN calculator
	/// </summary>
	public class ReversePolishNotationCalculator
	{
        // Memory to hold all the operands and operators.
		private readonly Stack<INode> _stack;
		
        // Evaluator to evaluate the expression.
		private readonly Evaluator _evaluator;

		/// <summary>
		/// Constructor for the calculator
		/// </summary>
		public ReversePolishNotationCalculator ()
		{
			_stack = new Stack<INode>();
			_evaluator = new Evaluator(_stack);
		}

		/// <summary>
		/// Is the calculator evaluation stack empty?
		/// </summary>
		public Boolean IsStackEmpty()
		{
			return _stack.Count == 0;				
		}	

		/// <summary>
        /// Push "+" to the stack
		/// </summary>
		public void PushPlusOperator()
		{
			_stack.Push(new PlusOperator());
		}

		/// <summary>
        /// Push "-" to the stack
		/// </summary>
		public void PushMinusOperator()
		{
			_stack.Push(new MinusOperator());
		}

		/// <summary>
		/// Push "*" to the stack
		/// </summary>
		public void PushTimesOperator()
		{
			_stack.Push(new TimesOperator());
		}

		/// <summary>
		/// Push "/" to the stack
		/// </summary>
		public void PushDivideByOperator()
		{
			_stack.Push(new DivideByOperator());
		}

		/// <summary>
		/// Push an operand on top of the stack
		/// </summary>
		/// <param name="op">
		/// A <see cref="System.Double"/>
		/// </param>
		public void PushOperand(double op)
		{
			_stack.Push(new Operand(op));
		}

		/// <summary>
		/// Pop the last operand from memory.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Double"/>
		/// </returns>
		public double PopOperand()
		{
		    if (_stack.Count == 0)
		    {
		        throw new EmptyStackEvaluationException("Stack is empty!");
		    }

		    INode res = _stack.Pop();


		    if (res as Operand != null)
		    {
		        return ((Operand) res).OperandValue;
		    }

		    throw new InvalidOperandRetrievalException(string.Format(@"Cannot pop an operand from a stack, top element is operator ""{0}""", res));
		}

		/// <summary>
		/// Peek the last operand in memory (don't remove it) 
		/// </summary>
		/// <returns>
		/// A <see cref="System.Double"/>
		/// </returns>
		public double PeekOperand()
		{
            if (_stack.Count == 0)
            {
                throw new EmptyStackEvaluationException("Stack is empty!");
            }

            INode res = _stack.Peek();


            if (res as Operand != null)
            {
                return ((Operand)res).OperandValue;
            }

            throw new InvalidOperandRetrievalException(string.Format(@"Cannot peek an operand from a stack, top element is operator ""{0}""", res));				
		}

        public void Clear()
        {
            _stack.Clear();
        }

		/// <summary>
		/// Evaluate the stack.
		/// </summary>
		public void Evaluate()
		{ 
			if (IsStackEmpty())
			{
				throw new EmptyStackEvaluationException("Unexpected empty stack. Cannot evaluate empty stack.");
			}

			INode node = _stack.Pop();
			node.Accept(_evaluator);
			_stack.Push(new Operand(_evaluator.Result));		
		}

        /// <summary>
        /// Get the string representation of the calculator memory.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            var printer = new NodePrinter();

            foreach (INode node in _stack)
            {
                node.Accept(printer);
            }
            return printer.Result.Trim();
        }
	}
}
