﻿using System;
using System.Collections.Generic;
using System.Resources;
using System.Globalization;
using System.Collections;
using ETHZ.C0284H.Calculator.Exceptions;
using ETHZ.C0284H.Calculator.Instruct;
using ETHZ.C0284H.Calculator.Instruct.Commands;
using ETHZ.C0284H.Calculator.Utils;

namespace ETHZ.C0284H.Calculator.UI
{
    sealed class UIResourceManager
    {
        // Singleton resource manager.
        private static readonly UIResourceManager s_manager;

        // Resource manager.
        private readonly ResourceManager _resourceManager;

        /// <summary>
        /// Object lock
        /// </summary>
        private readonly object _lock = new object();

        private LanguageInstruct _currentLanguage;

        /// <summary>
        /// Static constructor to create the singleton.
        /// </summary>
        static UIResourceManager()
        {
            s_manager = new UIResourceManager("ETHZ.C0284H.Calculator.Calculator");


        }

        public static UIResourceManager Instance
        {
            get { return s_manager; }
        }


        /// <summary>
        /// Get the name of current language.
        /// </summary>
        /// <value></value>
        public string CurrentLanguageName
        {
            get
            {
                lock (_lock)
                {
                    return _currentLanguage.InstructText;
                }
            }
        }

        public void SetLanguage(LanguageInstruct language)
        {
            if (language == null)
            {
                throw new ArgumentNullException("language");
            }

            lock (_lock)
            {
                _currentLanguage = language;
            }
        }

        /// <summary>
        /// Creates a new instance of ui resourse manager, sets current language to english.
        /// </summary>
        /// <param name="baseName"></param>
        private UIResourceManager(string baseName)
        {
            _resourceManager = new ResourceManager(baseName, GetType().Assembly);

            _currentLanguage = InstructRepository.English;
        }

        /// <summary>
        /// Current Culture
        /// </summary>
        public CultureInfo CurrentCulture 
        { 
            get
            {
                lock (_lock)
                {
                    return _currentLanguage.Culture;
                }
            }
        }

        /// <summary>
        /// Get localized message by id.
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public string GetLocalizedMessage(UIMessageID key)
        {
            return LocalizeString(GetInterfaceMessageIdAsString(key));
        }

        /// <summary>
        /// Get localized command text, according to the given type.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public string GetLocalizedCommand(CalculatorCommandType type)
        {
            return LocalizeString(GetCalculatorCommandAsString(type));
        }

        private string LocalizeString(string toLocalize)
        {
            return _resourceManager.GetString(toLocalize, CurrentCulture);
        }

        // Convert the message id into a string, which can be used to index the messages in resource file.
        private static string GetInterfaceMessageIdAsString(UIMessageID key)
        {
            return Enum.GetName(typeof(UIMessageID), key).ToLower();
        }

        // Get the command string.
        private static string GetCalculatorCommandAsString(CalculatorCommandType type)
        {
            return Enum.GetName(typeof(CalculatorCommandType), type).ToLower();
        }
    }
}
